@@ -1,7 +1,11 @@
-from functools import lru_cache, reduce
-from typing import List, Union, TextIO, Optional
-from linkml_runtime.utils.schemaview import SchemaView
+import logging
+import yaml
+import csv
 import deprecation
+import requests
+from functools import lru_cache, reduce
+from typing import List, Union, TextIO, Optional, Dict
+from linkml_runtime.utils.schemaview import SchemaView
 from linkml_runtime.linkml_model.meta import (
     SchemaDefinition,
     Element,
@@ -10,22 +14,26 @@ from linkml_runtime.linkml_model.meta import (
     ClassDefinition,
     SlotDefinition,
 )
+from pprint import pprint
 
-from bmt.toolkit_generator import ToolkitGenerator
 from bmt.utils import format_element, parse_name
-import logging
 
 Url = str
 Path = str
 
-REMOTE_PATH = (
-    "https://raw.githubusercontent.com/biolink/biolink-model/2.2.13/biolink-model.yaml"
-)
+REMOTE_PATH = "https://raw.githubusercontent.com/biolink/biolink-model/v3.1.2/biolink-model.yaml"
+PREDICATE_MAP = 'https://raw.githubusercontent.com/biolink/biolink-model/v3.1.2/predicate_mapping.yaml'
+INFORES_MAP = 'https://raw.githubusercontent.com/biolink/biolink-model/v3.1.2/infores_catalog_nodes.tsv'
+
+NODE_PROPERTY = "node property"
+ASSOCIATION_SLOT = "association slot"
 RELATED_TO = "related to"
+
 CACHE_SIZE = 1024
 
 logger = logging.getLogger(__name__)
 
+
 class Toolkit(object):
     """
     Provides a series of methods for performing lookups on the Biolink Model
@@ -38,11 +46,29 @@ class Toolkit(object):
     """
 
     def __init__(
-        self, schema: Union[Url, Path, TextIO, SchemaDefinition] = REMOTE_PATH
+            self, schema: Union[Url, Path, TextIO, SchemaDefinition] = REMOTE_PATH,
+            predicate_map: Url = PREDICATE_MAP,
+            infores_map: Url = INFORES_MAP
     ) -> None:
-        self.generator = ToolkitGenerator(schema)
-        self.generator.serialize()
         self.view = SchemaView(schema)
+        r = requests.get(predicate_map)
+        self.pmap = yaml.safe_load(r.text)
+
+        r = requests.get(infores_map)
+        content = r.content.decode('iso-8859-1')
+        self.infores_map = {}
+        for line in csv.reader(content.splitlines(), delimiter='\t'):
+            if line[2] == 'id':
+                continue
+            self.infores_map[line[2]] = {
+                "status": line[0],
+                "name": line[1],
+                "url": line[3],
+                "synonyms": line[4],
+                "has_contributor": line[6],
+                "description": line[7],
+                "category": line[8]
+            }
 
     @lru_cache(CACHE_SIZE)
     def get_all_elements(self, formatted: bool = False) -> List[str]:
@@ -89,7 +115,7 @@ class Toolkit(object):
 
         """
         classes = []
-        for x in self.generator.schema.classes:
+        for x in self.view.schema.classes:
             classes.append(x)
         filtered_classes = self._filter_secondary(classes)
         return self._format_all_elements(filtered_classes, formatted)
@@ -114,7 +140,7 @@ class Toolkit(object):
 
         """
         slots = []
-        for x in self.generator.schema.slots:
+        for x in self.view.schema.slots:
             slots.append(x)
         filtered_slots = self._filter_secondary(slots)
         return self._format_all_elements(filtered_slots, formatted)
@@ -139,7 +165,7 @@ class Toolkit(object):
 
         """
         types = []
-        for x in self.generator.schema.types:
+        for x in self.view.all_types():
             types.append(x)
         return self._format_all_elements(types, formatted)
 
@@ -255,7 +281,7 @@ class Toolkit(object):
         """
         filtered_elements = []
         for e in elements:
-            eo = self.generator.obj_for(e)
+            eo = self.view.get_element(e)
             if isinstance(eo, SlotDefinition):
                 if not eo.alias:
                     filtered_elements.append(e)
@@ -263,13 +289,115 @@ class Toolkit(object):
                 filtered_elements.append(e)
         return filtered_elements
 
+    @lru_cache(CACHE_SIZE)
+    def get_permissible_value_ancestors(self, permissible_value: str, enum_name: str, formatted: bool = False) -> List[
+        str]:
+        """
+        Get ancestors of a permissible value.
+
+        This method returns a list containing all the ancestors of a
+        permissible value of a given enum.
+
+        Parameters
+        ----------
+        enum_name: str
+            The name of the enum
+        permissible_value: str
+            The name of the permissible value
+
+        Returns
+        -------
+        List[str]
+            A list of elements
+
+        """
+        ancestors = self.view.permissible_value_ancestors(permissible_value, enum_name)
+        if formatted:
+            return self._format_all_elements(ancestors)
+        return ancestors
+
+    @lru_cache(CACHE_SIZE)
+    def get_infores_details(self, infores_id: str):
+        """
+        Get details of an information resource.
+
+        This method returns a dictionary containing details of a given
+        information resource.
+
+        Parameters
+        ----------
+        infores_id: str
+            The identifier of the information resource
+
+        Returns
+        -------
+        Dict[str, Any]
+            A dictionary containing details of the information resource
+
+        """
+        pprint(self.infores_map)
+        infores = self.infores_map.get(infores_id)
+        return infores
+
+    @lru_cache(CACHE_SIZE)
+    def get_predicate_mapping(self, mapped_predicate: str) -> Dict[str, str]:
+        """
+        Get the predicates that map to a given predicate.
+
+        This method returns a list containing all the predicates that map to
+        a given predicate.
+
+        Parameters
+        ----------
+        mapped_predicate: str
+            The name of the mapped predicate
+
+        Returns
+        -------
+        List[str]
+            A list of elements
+
+        """
+        association = {}
+
+        for mp in self.pmap.values():
+            for item in mp:
+                if item['mapped predicate'] == mapped_predicate:
+                    for k, v in item.items():
+                        association[format_element(self.get_element(k))] = v
+        return association
+
+    @lru_cache(CACHE_SIZE)
+    def get_permissible_value_parent(self, permissible_value: str, enum_name: str) -> str:
+        """
+        Get parent of a permissible value.
+
+        This method returns a list containing all the parent of a
+        permissible value of a given enum.
+
+        Parameters
+        ----------
+        enum_name: str
+            The name of the enum
+        permissible_value: str
+            The name of the permissible value
+
+        Returns
+        -------
+        List[str]
+            A list of elements
+
+        """
+        parent = self.view.permissible_value_parent(permissible_value, enum_name)
+        return parent
+
     @lru_cache(CACHE_SIZE)
     def get_ancestors(
-        self,
-        name: str,
-        reflexive: bool = True,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            name: str,
+            reflexive: bool = True,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Gets a list of names of ancestors.
@@ -309,17 +437,17 @@ class Toolkit(object):
             if a_element.mixins:
                 for mixin in a_element.mixins:
                     mixin_element = self.get_element(mixin)
-                    mixin_parents = self.generator.ancestors(mixin_element)
+                    mixin_parents = self.view.ancestors(mixin_element)
                     mixins_parents = mixins_parents + mixin_parents
         return mixins_parents
 
     @lru_cache(CACHE_SIZE)
     def get_descendants(
-        self,
-        name: str,
-        reflexive: bool = True,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            name: str,
+            reflexive: bool = True,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Gets a list of names of descendants.
@@ -358,9 +486,27 @@ class Toolkit(object):
 
         return self._format_all_elements(filtered_desc, formatted)
 
+    @lru_cache(CACHE_SIZE)
+    def get_all_multivalued_slots(self) -> List[str]:
+        """
+        Gets a list of names of all multivalued slots.
+
+        Returns
+        -------
+        List[str]
+            The names of all multivalued slots
+
+        """
+        multivalued_slots = []
+        slots = self.view.all_slots()
+        for slot_name, slot_def in slots.items():
+            if self.view.is_multivalued(slot_name):
+                multivalued_slots.append(slot_name)
+        return multivalued_slots
+
     @lru_cache(CACHE_SIZE)
     def get_children(
-        self, name: str, formatted: bool = False, mixin: bool = True
+            self, name: str, formatted: bool = False, mixin: bool = True
     ) -> List[str]:
         """
         Gets a list of names of children.
@@ -383,7 +529,7 @@ class Toolkit(object):
         children = []
         element = self.get_element(name)
         if element:
-            children = self.generator.children(element.name, mixin)
+            children = self.view.get_children(element.name, mixin)
         return self._format_all_elements(children, formatted)
 
     @lru_cache(CACHE_SIZE)
@@ -433,22 +579,25 @@ class Toolkit(object):
         """
         parsed_name = parse_name(name)
         logger.debug(parsed_name)
-        element = self.generator.obj_for(parsed_name)
-        if element is None and name in self.generator.aliases:
-            logger.debug("in aliases")
-            logger.debug(self.generator.aliases)
-            element = self.get_element(self.generator.aliases[name])
+        element = self.view.get_element(parsed_name)
+        if element is None and self.view.all_aliases() is not None:
+            for e in self.view.all_aliases():
+                if name in self.view.all_aliases()[e]:
+                    element = self.view.get_element(e)
         if element is None and "_" in name:
-            logger.debug("has a _")
             element = self.get_element(name.replace("_", " "))
+        if element is None:
+            for e, el in self.view.all_elements().items():
+                if el.name.lower() == name.lower():
+                    element = el
         return element
 
     def get_slot_domain(
-        self,
-        slot_name,
-        include_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            slot_name,
+            include_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Get the domain for a given slot.
@@ -492,11 +641,11 @@ class Toolkit(object):
         return self._format_all_elements(slot_domain, formatted)
 
     def get_slot_range(
-        self,
-        slot_name,
-        include_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            slot_name,
+            include_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Get the range for a given slot.
@@ -529,11 +678,11 @@ class Toolkit(object):
         return self._format_all_elements(slot_range, formatted)
 
     def get_all_slots_with_class_domain(
-        self,
-        class_name,
-        check_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            class_name,
+            check_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Given a class, get all the slots where the class is the domain.
@@ -561,11 +710,11 @@ class Toolkit(object):
         return self._format_all_elements(slot_names, formatted)
 
     def get_all_slots_with_class_range(
-        self,
-        class_name,
-        check_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            class_name,
+            check_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Given a class, get all the slots where the class is the range.
@@ -593,11 +742,11 @@ class Toolkit(object):
         return self._format_all_elements(slot_names, formatted)
 
     def get_all_predicates_with_class_domain(
-        self,
-        class_name,
-        check_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            class_name,
+            check_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Given a class, get all Biolink predicates where the class is the domain.
@@ -631,11 +780,11 @@ class Toolkit(object):
         return self._format_all_elements(filtered_slots, formatted)
 
     def get_all_predicates_with_class_range(
-        self,
-        class_name,
-        check_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            class_name,
+            check_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ):
         """
         Given a class, get all Biolink predicates where the class is the range.
@@ -669,11 +818,11 @@ class Toolkit(object):
         return self._format_all_elements(filtered_slots, formatted)
 
     def get_all_properties_with_class_domain(
-        self,
-        class_name,
-        check_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            class_name,
+            check_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Given a class, get all Biolink properties where the class is the domain.
@@ -707,11 +856,11 @@ class Toolkit(object):
         return self._format_all_elements(filtered_slots, formatted)
 
     def get_all_properties_with_class_range(
-        self,
-        class_name,
-        check_ancestors: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            class_name,
+            check_ancestors: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> List[str]:
         """
         Given a class, get all Biolink properties where the class is the range.
@@ -765,18 +914,20 @@ class Toolkit(object):
         element = self.get_element(slot_name)
         if element:
             types = self.get_all_types()
+            if element.range is None and self.view.schema.default_range:
+                element.range = self.view.schema.default_range
             if element.range in types:
                 et = element.range
             else:
                 et = "uriorcurie"
             if formatted:
-                element_type = format_element(self.generator.obj_for(et))
+                element_type = format_element(self.view.get_element(et))
             else:
                 element_type = et
         return element_type
 
     def _get_all_slots_with_class_domain(
-        self, element: Element, check_ancestors: bool, mixin: bool = True
+            self, element: Element, check_ancestors: bool, mixin: bool = True
     ) -> List[Element]:
         """
         Given a class, get all the slots where the class is the domain.
@@ -797,7 +948,7 @@ class Toolkit(object):
 
         """
         slots = []
-        for k, v in self.generator.schema.slots.items():
+        for k, v in self.view.schema.slots.items():
             if check_ancestors:
                 if (v.domain == element.name or v.domain in self.get_ancestors(element.name, mixin)
                         or element.name in v.domain_of
@@ -809,7 +960,7 @@ class Toolkit(object):
         return slots
 
     def _get_all_slots_with_class_range(
-        self, element: Element, check_ancestors: bool, mixin: bool = True
+            self, element: Element, check_ancestors: bool, mixin: bool = True
     ) -> List[Element]:
         """
         Given a class, get all the slots where the class is the range.
@@ -830,10 +981,10 @@ class Toolkit(object):
 
         """
         slots = []
-        for k, v in self.generator.schema.slots.items():
+        for k, v in self.view.schema.slots.items():
             if check_ancestors:
                 if v.range == element.name or v.range in self.get_ancestors(
-                    element.name, mixin
+                        element.name, mixin
                 ):
                     slots.append(v)
             else:
@@ -841,10 +992,52 @@ class Toolkit(object):
                     slots.append(v)
         return slots
 
+    @lru_cache(CACHE_SIZE)
+    def is_node_property(self, name: str, mixin: bool = True) -> bool:
+        """
+        Determines whether the given name is the name of a node property
+        in the Biolink Model. An element is a node property if it descends from
+        `NODE_PROPERTY`
+
+        Parameters
+        ----------
+        name: str
+            The name or alias of an element in the Biolink Model
+        mixin: bool
+            If True, then that means we want to find mixin ancestors as well as is_a ancestors
+
+        Returns
+        -------
+        bool
+            That the named element is a valid node property in Biolink Model
+        """
+        return NODE_PROPERTY in self.get_ancestors(name, mixin)
+
+    @lru_cache(CACHE_SIZE)
+    def is_association_slot(self, name: str, mixin: bool = True) -> bool:
+        """
+        Determines whether the given name is the name of an association slot
+        in the Biolink Model. An element is an association slot if it descends from
+        `ASSOCIATION_SLOT`
+
+        Parameters
+        ----------
+        name: str
+            The name or alias of an element in the Biolink Model
+        mixin: bool
+            If True, then that means we want to find mixin ancestors as well as is_a ancestors
+
+        Returns
+        -------
+        bool
+            That the named element is a valid an association slot in Biolink Model
+        """
+        return ASSOCIATION_SLOT in self.get_ancestors(name, mixin)
+
     @lru_cache(CACHE_SIZE)
     def is_predicate(self, name: str, mixin: bool = True) -> bool:
         """
-        Determines whether the given name is the name of an relation/predicate
+        Determines whether the given name is the name of a relation/predicate
         in the Biolink Model. An element is a predicate if it descends from
         `RELATED_TO`
 
@@ -888,7 +1081,7 @@ class Toolkit(object):
                 annotation_tags.append(annotation)
         is_canonical = (
             True
-            if element is not None and "biolink:canonical_predicate" in annotation_tags
+            if element is not None and "canonical_predicate" in annotation_tags
             else False
         )
         return (
@@ -962,7 +1155,7 @@ class Toolkit(object):
 
         """
         parsed_name = parse_name(name)
-        element = self.generator.obj_for(parsed_name)
+        element = self.view.get_element(parsed_name)
         return subset in element.in_subset
 
     @lru_cache(CACHE_SIZE)
@@ -1023,11 +1216,11 @@ class Toolkit(object):
 
     @lru_cache(CACHE_SIZE)
     def get_element_by_mapping(
-        self,
-        identifier: str,
-        most_specific: bool = False,
-        formatted: bool = False,
-        mixin: bool = True,
+            self,
+            identifier: str,
+            most_specific: bool = False,
+            formatted: bool = False,
+            mixin: bool = True,
     ) -> Optional[str]:
         """
         Get a Biolink Model element by mapping.
@@ -1073,7 +1266,7 @@ class Toolkit(object):
                 logger.debug(a)
                 if a in common_ancestors:
                     if formatted:
-                        element = format_element(self.generator.obj_for(a))
+                        element = format_element(self.view.get_element(a))
                     else:
                         element = a
                     return element
@@ -1098,8 +1291,8 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
+        mappings = self.view.get_mappings().get(
+            self.view.namespaces.uri_for(identifier), set()
         )
         if not mappings:
             exact = set(self.get_element_by_exact_mapping(identifier))
@@ -1120,7 +1313,7 @@ class Toolkit(object):
 
     @lru_cache(CACHE_SIZE)
     def get_element_by_exact_mapping(
-        self, identifier: str, formatted: bool = False
+            self, identifier: str, formatted: bool = False
     ) -> List[str]:
         """
         Given an identifier as IRI/CURIE, find a Biolink element that corresponds
@@ -1139,15 +1332,15 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.exact_mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
+        mappings = self.view.exact_mappings.get(
+            self.view.namespaces.uri_for(identifier), set()
         )
         logger.debug(mappings)
         return self._format_all_elements(mappings, formatted)
 
     @lru_cache(CACHE_SIZE)
     def get_element_by_close_mapping(
-        self, identifier: str, formatted: bool = False
+            self, identifier: str, formatted: bool = False
     ) -> List[str]:
         """
         Given an identifier as IRI/CURIE, find a Biolink element that corresponds
@@ -1166,14 +1359,14 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.close_mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
+        mappings = self.view.close_mappings.get(
+            self.view.namespaces.uri_for(identifier), set()
         )
         return self._format_all_elements(mappings, formatted)
 
     @lru_cache(CACHE_SIZE)
     def get_element_by_related_mapping(
-        self, identifier: str, formatted: bool = False
+            self, identifier: str, formatted: bool = False
     ) -> List[str]:
         """
         Given an identifier as IRI/CURIE, find a Biolink element that corresponds
@@ -1192,14 +1385,14 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.related_mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
+        mappings = self.view.related_mappings.get(
+            self.view.namespaces.uri_for(identifier), set()
         )
         return self._format_all_elements(mappings, formatted)
 
     @lru_cache(CACHE_SIZE)
     def get_element_by_narrow_mapping(
-        self, identifier: str, formatted: bool = False
+            self, identifier: str, formatted: bool = False
     ) -> List[str]:
         """
         Given an identifier as IRI/CURIE, find a Biolink element that corresponds
@@ -1218,14 +1411,14 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.narrow_mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
+        mappings = self.view.narrow_mappings.get(
+            self.view.namespaces.uri_for(identifier), set()
         )
         return self._format_all_elements(mappings, formatted)
 
     @lru_cache(CACHE_SIZE)
     def get_element_by_broad_mapping(
-        self, identifier: str, formatted: bool = False
+            self, identifier: str, formatted: bool = False
     ) -> List[str]:
         """
         Given an identifier as IRI/CURIE, find a Biolink element that corresponds
@@ -1244,17 +1437,17 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.broad_mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
+        mappings = self.view.broad_mappings.get(
+            self.view.namespaces.uri_for(identifier), set()
         )
         return self._format_all_elements(mappings, formatted)
 
     @lru_cache(CACHE_SIZE)
     def get_all_elements_by_mapping(
-        self, identifier: str, formatted: bool = False
+            self, identifier: str, formatted: bool = False
     ) -> List[str]:
         """
-        Given an identifier as IRI/CURIE, find all Biolink element that corresponds
+        Given an identifier as IRI/CURIE, find all Biolink elements that correspond
         to the given identifier as part of its mappings.
 
         Parameters
@@ -1270,23 +1463,11 @@ class Toolkit(object):
             A list of Biolink elements that correspond to the given identifier IRI/CURIE
 
         """
-        mappings = self.generator.mappings.get(
-            self.generator.namespaces.uri_for(identifier), set()
-        )
-        exact = set(self.get_element_by_exact_mapping(identifier))
-        mappings.update(exact)
-        close = set(self.get_element_by_close_mapping(identifier))
-        mappings.update(close)
-        related = set(self.get_element_by_related_mapping(identifier))
-        mappings.update(related)
-        narrow = set(self.get_element_by_narrow_mapping(identifier))
-        mappings.update(narrow)
-        broad = set(self.get_element_by_broad_mapping(identifier))
-        mappings.update(broad)
+        mappings = self.view.get_element_by_mapping(identifier)
         return self._format_all_elements(mappings, formatted)
 
     def _format_all_elements(
-        self, elements: List[str], formatted: bool = False
+            self, elements: List[str], formatted: bool = False
     ) -> List[str]:
         """
         Format all the elements in a given list.
@@ -1306,7 +1487,7 @@ class Toolkit(object):
         """
         if formatted:
             formatted_elements = [
-                format_element(self.generator.obj_for(x)) for x in elements
+                format_element(self.view.get_element(x)) for x in elements
             ]
         else:
             formatted_elements = elements
@@ -1323,7 +1504,7 @@ class Toolkit(object):
             The biolink-model version
 
         """
-        return self.generator.schema.version
+        return self.view.schema.version
 
     @deprecation.deprecated(
         deprecated_in="0.3.0",
